#
# DeepLabCut Toolbox (deeplabcut.org)
# © A. & M.W. Mathis Labs
# https://github.com/DeepLabCut/DeepLabCut
#
# Please see AUTHORS for contributors.
# https://github.com/DeepLabCut/DeepLabCut/blob/master/AUTHORS
#
# Licensed under GNU Lesser General Public License v3.0
#
""" Cosine Scheduler

Cosine LR schedule with warmup, cycle/restarts, noise.

Hacked together by / Copyright 2020 Ross Wightman
"""
import logging
import math
import torch

from .scheduler import Scheduler


_logger = logging.getLogger(__name__)


class CosineLRScheduler(Scheduler):
    """
    Cosine decay with restarts.
    This is described in the paper https://arxiv.org/abs/1608.03983.

    Inspiration from
    https://github.com/allenai/allennlp/blob/master/allennlp/training/learning_rate_schedulers/cosine.py
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        t_initial: int,
        t_mul: float = 1.0,
        lr_min: float = 0.0,
        decay_rate: float = 1.0,
        warmup_t=0,
        warmup_lr_init=0,
        warmup_prefix=False,
        cycle_limit=0,
        t_in_epochs=True,
        noise_range_t=None,
        noise_pct=0.67,
        noise_std=1.0,
        noise_seed=42,
        initialize=True,
    ) -> None:
        super().__init__(
            optimizer,
            param_group_field="lr",
            noise_range_t=noise_range_t,
            noise_pct=noise_pct,
            noise_std=noise_std,
            noise_seed=noise_seed,
            initialize=initialize,
        )

        assert t_initial > 0
        assert lr_min >= 0
        if t_initial == 1 and t_mul == 1 and decay_rate == 1:
            _logger.warning(
                "Cosine annealing scheduler will have no effect on the learning "
                "rate since t_initial = t_mul = eta_mul = 1."
            )
        self.t_initial = t_initial
        self.t_mul = t_mul
        self.lr_min = lr_min
        self.decay_rate = decay_rate
        self.cycle_limit = cycle_limit
        self.warmup_t = warmup_t
        self.warmup_lr_init = warmup_lr_init
        self.warmup_prefix = warmup_prefix
        self.t_in_epochs = t_in_epochs
        if self.warmup_t:
            self.warmup_steps = [
                (v - warmup_lr_init) / self.warmup_t for v in self.base_values
            ]
            super().update_groups(self.warmup_lr_init)
        else:
            self.warmup_steps = [1 for _ in self.base_values]

    def _get_lr(self, t):
        if t < self.warmup_t:
            lrs = [self.warmup_lr_init + t * s for s in self.warmup_steps]
        else:
            if self.warmup_prefix:
                t = t - self.warmup_t

            if self.t_mul != 1:
                i = math.floor(
                    math.log(1 - t / self.t_initial * (1 - self.t_mul), self.t_mul)
                )
                t_i = self.t_mul**i * self.t_initial
                t_curr = t - (1 - self.t_mul**i) / (1 - self.t_mul) * self.t_initial
            else:
                i = t // self.t_initial
                t_i = self.t_initial
                t_curr = t - (self.t_initial * i)

            gamma = self.decay_rate**i
            lr_min = self.lr_min * gamma
            lr_max_values = [v * gamma for v in self.base_values]

            if self.cycle_limit == 0 or (self.cycle_limit > 0 and i < self.cycle_limit):
                lrs = [
                    lr_min
                    + 0.5 * (lr_max - lr_min) * (1 + math.cos(math.pi * t_curr / t_i))
                    for lr_max in lr_max_values
                ]
            else:
                lrs = [self.lr_min for _ in self.base_values]

        return lrs

    def get_epoch_values(self, epoch: int):
        if self.t_in_epochs:
            return self._get_lr(epoch)
        else:
            return None

    def get_update_values(self, num_updates: int):
        if not self.t_in_epochs:
            return self._get_lr(num_updates)
        else:
            return None

    def get_cycle_length(self, cycles=0):
        if not cycles:
            cycles = self.cycle_limit
        cycles = max(1, cycles)
        if self.t_mul == 1.0:
            return self.t_initial * cycles
        else:
            return int(
                math.floor(
                    -self.t_initial * (self.t_mul**cycles - 1) / (1 - self.t_mul)
                )
            )



#
# DeepLabCut Toolbox (deeplabcut.org)
# © A. & M.W. Mathis Labs
# https://github.com/DeepLabCut/DeepLabCut
#
# Please see AUTHORS for contributors.
# https://github.com/DeepLabCut/DeepLabCut/blob/main/AUTHORS
#
# Licensed under GNU Lesser General Public License v3.0
#
from deeplabcut.pose_estimation_pytorch.post_processing.match_predictions_to_gt import (
    oks_match_prediction_to_gt,
    rmse_match_prediction_to_gt,
)



#
# DeepLabCut Toolbox (deeplabcut.org)
# © A. & M.W. Mathis Labs
# https://github.com/DeepLabCut/DeepLabCut
#
# Please see AUTHORS for contributors.
# https://github.com/DeepLabCut/DeepLabCut/blob/master/AUTHORS
#
# Licensed under GNU Lesser General Public License v3.0
#
from __future__ import annotations

import os

from PySide6 import QtWidgets
from PySide6.QtCore import Qt, Slot
from PySide6.QtGui import QIcon

from deeplabcut.core.config import read_config_as_dict
from deeplabcut.gui.dlc_params import DLCParams
from deeplabcut.gui.widgets import ConfigEditor
from pathlib import Path


def _create_label_widget(
    text: str,
    style: str = "",
    margins: tuple = (20, 10, 0, 10),
) -> QtWidgets.QLabel:
    label = QtWidgets.QLabel(text)
    label.setContentsMargins(*margins)
    label.setStyleSheet(style)

    return label


def _create_horizontal_layout(
    alignment=None, spacing: int = 20, margins: tuple = (20, 0, 0, 0)
) -> QtWidgets.QHBoxLayout():
    layout = QtWidgets.QHBoxLayout()
    layout.setAlignment(Qt.AlignLeft | Qt.AlignTop)
    layout.setSpacing(spacing)
    layout.setContentsMargins(*margins)

    return layout


def _create_vertical_layout(
    alignment=None, spacing: int = 20, margins: tuple = (20, 0, 0, 0)
) -> QtWidgets.QVBoxLayout():
    layout = QtWidgets.QVBoxLayout()
    layout.setAlignment(Qt.AlignLeft | Qt.AlignTop)
    layout.setSpacing(spacing)
    layout.setContentsMargins(*margins)

    return layout


def _create_grid_layout(
    alignment=None,
    spacing: int = 20,
    margins: tuple = None,
) -> QtWidgets.QGridLayout:
    layout = QtWidgets.QGridLayout()
    layout.setAlignment(Qt.AlignLeft | Qt.AlignTop)
    layout.setSpacing(spacing)
    if margins:
        layout.setContentsMargins(*margins)

    return layout


def set_combo_items(combo_box: QtWidgets.QComboBox, items: list[str], index: int = 0):
    """
    Safely replaces all items in a QComboBox and sets the current index,
    ensuring that the `currentTextChanged` signal is emitted exactly once
    (and only if items are present).

    This method suppresses intermediate signal emissions that can be triggered
    by `clear()` and `addItems()` — both of which may emit multiple signals
    depending on the underlying Qt model and signal connections.

    It also handles the edge case where the item at the target index is already
    selected: by default, Qt will not emit a signal if the index doesn't change.
    To ensure consistent behavior, this method temporarily sets the index to -1
    (i.e., no selection), which is done with signals blocked, then restores the
    intended index — causing the signal to emit once and only once.

    Parameters:
        combo_box (QComboBox): The combo box to update.
        items (list of str): New items to populate the combo box.
        index (int): The index to select after updating items. Defaults to 0.

    Note:
        - If the items list is empty, no item will be selected and no signal will be emitted.
        - This method is designed to be safe for use with PySide, where signals
          cannot be manually emitted, and future-proof if multiple slots are connected.
    """
    combo_box.blockSignals(True)
    combo_box.clear()
    combo_box.addItems(items)
    combo_box.blockSignals(False)

    if not items:
        combo_box.setCurrentIndex(-1)
        return

    current = combo_box.currentIndex()
    if current == index:
        # Temporarily change index to suppress duplicate signal
        combo_box.blockSignals(True)
        combo_box.setCurrentIndex(-1)
        combo_box.blockSignals(False)

    combo_box.setCurrentIndex(index)


class BodypartListWidget(QtWidgets.QListWidget):
    def __init__(
        self,
        root: QtWidgets.QMainWindow,
        parent: QtWidgets.QWidget,
        # all_bodyparts: List
        # NOTE: Is there a case where a specific list should
        # have bodyparts other than the root? I don't think so.
    ):
        super().__init__()

        self.root = root
        self.parent = parent
        self.selected_bodyparts = self.root.all_bodyparts

        self.setEnabled(False)
        self.setMaximumWidth(600)
        self.setMaximumHeight(500)
        self.hide()

        self.addItems(self.root.all_bodyparts)
        self.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)

        self.itemSelectionChanged.connect(self.update_selected_bodyparts)

    def refresh(self):
        self.clear()
        self.addItems(self.root.all_bodyparts)
        self.update_selected_bodyparts()

    def update_selected_bodyparts(self):
        self.selected_bodyparts = [item.text() for item in self.selectedItems()]
        self.root.logger.info(f"Selected bodyparts:\n\t{self.selected_bodyparts}")


class VideoSelectionWidget(QtWidgets.QWidget):
    def __init__(
        self,
        root: QtWidgets.QMainWindow,
        parent: QtWidgets.QWidget,
        hide_videotype: bool = False,
    ):
        super().__init__(parent)

        self.root = root
        self.parent = parent

        self._init_layout(hide_videotype)

    def _init_layout(self, hide_videotype: bool):
        layout = _create_horizontal_layout()

        # Videotype selection
        self.videotype_widget = QtWidgets.QComboBox()
        self.videotype_widget.setMinimumWidth(100)
        self.videotype_widget.addItems(DLCParams.VIDEOTYPES)
        self.videotype_widget.setCurrentText(self.root.video_type)
        self.root.video_type_.connect(self.videotype_widget.setCurrentText)
        self.videotype_widget.currentTextChanged.connect(self.update_videotype)

        # Select videos
        self.select_video_button = QtWidgets.QPushButton("Select videos")
        self.select_video_button.setMaximumWidth(200)
        self.select_video_button.clicked.connect(self.update_videos)
        self.root.video_files_.connect(self._update_video_selection)

        # Number of selected videos text
        self.selected_videos_text = QtWidgets.QLabel(
            ""
        )  # updated when videos are selected

        # Clear video selection
        self.clear_videos = QtWidgets.QPushButton("Clear selection")
        self.clear_videos.clicked.connect(self.clear_selected_videos)

        if not hide_videotype:
            layout.addWidget(self.videotype_widget)
        layout.addWidget(self.select_video_button)
        layout.addWidget(self.selected_videos_text)
        layout.addWidget(self.clear_videos, alignment=Qt.AlignRight)

        self.setLayout(layout)

    @property
    def files(self):
        return self.root.video_files

    def update_videotype(self, vtype):
        self.clear_selected_videos()
        self.root.video_type = vtype

    def _update_video_selection(self, videopaths):
        n_videos = len(self.root.video_files)
        if n_videos:
            self.selected_videos_text.setText(f"{n_videos} videos selected")
            self.select_video_button.setText("Add more videos")
        else:
            self.selected_videos_text.setText("")
            self.select_video_button.setText("Select videos")

    def update_videos(self):
        directory_to_open = self.root.project_folder

        # Create a filter string with both lowercase and uppercase extensions

        video_types = [f"*.{ext.lower()}" for ext in DLCParams.VIDEOTYPES[1:]] + [
            f"*.{ext.upper()}" for ext in DLCParams.VIDEOTYPES[1:]
        ]
        video_filter = f"Videos ({' '.join(video_types)})"

        filenames = QtWidgets.QFileDialog.getOpenFileNames(
            parent=self,
            caption="Select video(s) to analyze",
            dir=directory_to_open,
            filter=video_filter,
        )

        if filenames[0]:
            # Qt returns a tuple (list of files, filetype)
            self.root.add_video_files([os.path.abspath(vid) for vid in filenames[0]])

    def clear_selected_videos(self):
        self.root.clear_video_files()
        self.root.logger.info(f"Cleared selected videos")


class SnapshotSelectionWidget(QtWidgets.QWidget):
    def __init__(
        self,
        root: QtWidgets.QMainWindow,
        parent: QtWidgets.QWidget,
        margins: tuple,
        select_button_text: str,
    ):
        super().__init__(parent)
        self.root = root
        self.parent = parent
        self.selected_snapshot = None
        self._init_layout(margins, select_button_text)

    def _init_layout(self, margins, select_button_text):
        layout = _create_horizontal_layout(margins=margins)

        # Select snapshot
        self.select_snapshot_button = QtWidgets.QPushButton(select_button_text)
        self.select_snapshot_button.setMaximumWidth(200)
        self.select_snapshot_button.clicked.connect(self.select_snapshot)

        # Selected snapshot text
        self.selected_snapshot_text = QtWidgets.QLabel(
            ""
        )  # updated when snapshot is selected

        # Clear snapshot selection
        self.clear_snapshot_button = QtWidgets.QPushButton("Clear selection")
        self.clear_snapshot_button.clicked.connect(self.clear_selected_snapshot)
        self.clear_snapshot_button.hide()

        layout.addWidget(self.select_snapshot_button)
        layout.addWidget(self.selected_snapshot_text)
        layout.addWidget(self.clear_snapshot_button, alignment=Qt.AlignRight)

        self.setLayout(layout)

    def _update_selected_snapshot_display(self):
        if self.selected_snapshot is None:
            self.selected_snapshot_text.setText("")
            self.clear_snapshot_button.hide()
        else:
            self.selected_snapshot_text.setText(
                f"{os.path.basename(self.selected_snapshot)}"
            )
            self.clear_snapshot_button.show()

    def select_snapshot(self):
        # Create a filter string with both lowercase and uppercase extensions
        snapshot_types = ["*.pt", "*.PT"]
        snapshot_filter = f"Snapshots ({' '.join(snapshot_types)})"

        directory_to_open = self.root.models_folder

        selected_snapshot, _ = QtWidgets.QFileDialog.getOpenFileName(
            parent=self,
            caption="Select snapshot to start training from",
            dir=directory_to_open,
            filter=snapshot_filter,
        )
        # When Canceling a file selection, Qt returns an empty string as selected file
        if selected_snapshot:
            self.selected_snapshot = os.path.abspath(selected_snapshot)

        self._update_selected_snapshot_display()

    def clear_selected_snapshot(self):
        self.selected_snapshot = None
        self._update_selected_snapshot_display()


class ConditionsSelectionWidget(QtWidgets.QWidget):
    def __init__(
        self,
        root: QtWidgets.QMainWindow,
        parent: QtWidgets.QWidget,
    ):
        super().__init__(parent=parent)
        self.root = root
        self.parent = parent
        self.selected_conditions = None
        self._init_layout()

    def _init_layout(self):
        layout = _create_horizontal_layout()

        # Select conditions
        self.select_conditions_button = QtWidgets.QPushButton("Select conditions")
        self.select_conditions_button.setMaximumWidth(200)
        self.select_conditions_button.clicked.connect(self.select_conditions)

        # Selected conditions text
        self.selected_conditions_text = QtWidgets.QLabel(
            ""
        )  # updated when conditions are selected

        layout.addWidget(self.select_conditions_button)
        layout.addWidget(self.selected_conditions_text)

        self.setLayout(layout)

    def _update_selected_conditions_display(self):
        def _shorten_path(path: str, max_length: int = 30) -> str:
            if len(path) <= max_length:
                return path
            return "..." + path[-(max_length - 3) :]

        self.selected_conditions_text.setText(
            ""
            if self.selected_conditions is None
            else f"{_shorten_path(self.selected_conditions)}"
        )

    def select_conditions(self):
        def _is_model_bu(selected_conditions) -> bool:
            model_config_path = Path(selected_conditions).parent / "pytorch_config.yaml"
            model_config = read_config_as_dict(model_config_path)
            return model_config.get("method").lower() == "bu"

        # Create a filter string with both lowercase and uppercase extensions
        snapshots_label = "Snapshots"
        h5_predictions_label = "H5 predictions"
        json_prediction_label = "Json predictions"
        snapshot_types = ["*.pt", "*.PT"]
        h5_predictions_types = ["*.h5", "*.H5"]
        json_prediction_types = ["*.json", "*.JSON"]
        conditions_filter = ";;".join(
            [
                f"{snapshots_label} ({' '.join(snapshot_types)})",
                f"{h5_predictions_label} ({' '.join(h5_predictions_types)})",
                f"{json_prediction_label} ({' '.join(json_prediction_types)})",
            ]
        )

        directory_to_open = self.root.project_folder

        selected_conditions, selected_filter = QtWidgets.QFileDialog.getOpenFileName(
            parent=self,
            caption="Select conditions to use during inference (snapshot or predictions file)",
            dir=directory_to_open,
            filter=conditions_filter,
        )
        if selected_filter.startswith(snapshots_label) and selected_conditions:
            if not _is_model_bu(selected_conditions):
                msg = _create_message_box(
                    f"Invalid conditions",
                    (
                        f"The selected snapshot ({selected_conditions}) cannot be "
                        "used as conditions because it is not a Bottom-Up model."
                    ),
                )
                msg.exec_()
                selected_conditions = None

        # When Canceling a file selection, Qt returns an empty string as selected file
        self.selected_conditions = (
            str(os.path.abspath(selected_conditions)) if selected_conditions else None
        )

        self._update_selected_conditions_display()


class TrainingSetSpinBox(QtWidgets.QSpinBox):
    def __init__(self, root, parent):
        super().__init__(parent)

        self.root = root
        self.parent = parent

        self.setMaximum(100)
        self.setValue(self.root.trainingset_index)
        self.valueChanged.connect(self.root.update_trainingset)


class ShuffleSpinBox(QtWidgets.QSpinBox):
    def __init__(self, root, parent):
        super().__init__(parent)

        self.root = root
        self.parent = parent

        self.setMaximum(10_000)
        self.setValue(self.root.shuffle_value)
        self.valueChanged.connect(self.root.update_shuffle)
        self.root.shuffle_change.connect(self.update_shuffle)

    @Slot(int)
    def update_shuffle(self, new_shuffle: int):
        if new_shuffle != self.value():
            self.setValue(new_shuffle)


class DefaultTab(QtWidgets.QWidget):
    def __init__(
        self,
        root: QtWidgets.QMainWindow,
        parent: QtWidgets.QWidget = None,
        h1_description: str = "",
    ):
        super().__init__(parent)

        self.parent = parent
        self.root = root

        self.h1_description = h1_description

        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        self.setLayout(self.main_layout)

        self._init_default_layout()

    def _init_default_layout(self):
        # Add tab header
        self.main_layout.addWidget(
            _create_label_widget(self.h1_description, "font:bold;", (10, 10, 0, 10))
        )

        # Add separating line
        self.separator = QtWidgets.QFrame()
        self.separator.setFrameShape(QtWidgets.QFrame.HLine)
        self.separator.setFrameShadow(QtWidgets.QFrame.Raised)
        self.separator.setLineWidth(0)
        self.separator.setMidLineWidth(1)
        policy = QtWidgets.QSizePolicy()
        policy.setVerticalPolicy(QtWidgets.QSizePolicy.Policy.Fixed)
        policy.setHorizontalPolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        self.separator.setSizePolicy(policy)
        self.main_layout.addWidget(self.separator)


class EditYamlButton(QtWidgets.QPushButton):
    def __init__(
        self, button_label: str, filepath: str, parent: QtWidgets.QWidget = None
    ):
        super().__init__(parent)
        self.filepath = filepath
        self.parent = parent

        self.clicked.connect(self.open_config)

    def open_config(self):
        editor = ConfigEditor(self.filepath)
        editor.show()


class BrowseFilesButton(QtWidgets.QPushButton):
    def __init__(
        self,
        button_label: str,
        filetype: str = None,
        cwd: str = None,
        single_file: bool = False,
        dialog_text: str = None,
        file_text: str = None,
        parent=None,
    ):
        super().__init__(parent)
        self.filetype = filetype
        self.single_file_only = single_file
        self.cwd = cwd
        self.parent = parent

        self.dialog_text = dialog_text
        self.file_text = file_text

        self.files = set()

        self.clicked.connect(self.browse_files)

    def browse_files(self):
        # Look for any extension by default
        file_ext = "*"
        if self.filetype:
            # This works both with e.g. .avi and avi
            file_ext = self.filetype.split(".")[-1]

        # Choose multiple files by default
        open_file_func = QtWidgets.QFileDialog.getOpenFileNames
        if self.single_file_only:
            open_file_func = QtWidgets.QFileDialog.getOpenFileName

        cwd = ""
        if self.cwd:
            cwd = self.cwd

        dialog_text = f"Select .{file_ext} files"
        if self.dialog_text:
            dialog_text = self.dialog_text

        file_text = f"Files (*.{file_ext})"
        if self.file_text:
            file_text = self.file_text

        filepaths = open_file_func(self, dialog_text, cwd, file_text)

        if filepaths:
            self.files.update(filepaths[0])


def _create_message_box(text, info_text):
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Information)
    msg.setText(text)
    msg.setInformativeText(info_text)

    msg.setWindowTitle("Info")
    msg.setMinimumWidth(900)
    logo_dir = os.path.dirname(os.path.realpath("logo.png")) + os.path.sep
    logo = logo_dir + "/assets/logo.png"
    msg.setWindowIcon(QIcon(logo))
    msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
    return msg


def _create_confirmation_box(title, description):
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Information)
    msg.setText(title)
    msg.setInformativeText(description)

    msg.setWindowTitle("Confirmation")
    msg.setMinimumWidth(900)
    logo_dir = os.path.dirname(os.path.realpath("logo.png")) + os.path.sep
    logo = logo_dir + "/assets/logo.png"
    msg.setWindowIcon(QIcon(logo))
    msg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
    return msg


def set_layout_contents_visible(layout: QtWidgets.QLayout, visible: bool):
    for i in range(layout.count()):
        item = layout.itemAt(i)

        # If it's a widget item
        widget = item.widget()
        if widget is not None:
            widget.setVisible(visible)

        # If it's a nested layout
        child_layout = item.layout()
        if child_layout is not None:
            set_layout_contents_visible(child_layout, visible)



import torch
from torch import nn


class FilteredDetector(nn.Module):
    def __init__(self, base_model: nn.Module, class_id: int):
        """
        Wrap a torchvision detector to return predictions only for a single class.

        Args:
            base_model: A torchvision-style object detector.
            class_id: The integer class ID to keep (e.g., 1 for 'person' in COCO).
        """
        super().__init__()
        self.base_model = base_model
        self.class_id = class_id

    def forward(self, images: list[torch.Tensor]) -> list[dict[str, torch.Tensor]]:
        """
        Arguments:
            images: list of input images as Tensors

        Returns:
            List of dicts, each containing boxes/scores/labels filtered to the specified class.
        """
        with torch.no_grad():
            outputs = self.base_model(images)

        filtered_outputs = []
        for output in outputs:
            mask = output["labels"] == self.class_id
            filtered_output = {
                "boxes": output["boxes"][mask],
                "scores": output["scores"][mask],
                "labels": output["labels"][mask],
            }
            filtered_outputs.append(filtered_output)

        losses = {}
        return losses, filtered_outputs



#
# DeepLabCut Toolbox (deeplabcut.org)
# © A. & M.W. Mathis Labs
# https://github.com/DeepLabCut/DeepLabCut
#
# Please see AUTHORS for contributors.
# https://github.com/DeepLabCut/DeepLabCut/blob/master/AUTHORS
#
# Licensed under GNU Lesser General Public License v3.0
#
import os
import webbrowser
from functools import partial
from pathlib import Path

import dlclibrary
from PySide6 import QtWidgets
from PySide6.QtCore import QRegularExpression, Qt, QTimer, Signal, Slot, QSize
from PySide6.QtGui import QIcon, QPixmap, QRegularExpressionValidator
import cv2
import torch

import deeplabcut
from deeplabcut.core.engine import Engine
from deeplabcut.gui import BASE_DIR
from deeplabcut.gui.components import (
    _create_grid_layout,
    _create_label_widget,
    DefaultTab,
    VideoSelectionWidget,
    set_layout_contents_visible,
    set_combo_items,
)
from deeplabcut.gui.utils import move_to_separate_thread
from deeplabcut.gui.widgets import ClickableLabel
from deeplabcut.pose_estimation_pytorch.apis.utils import TORCHVISION_DETECTORS


class RegExpValidator(QRegularExpressionValidator):
    validationChanged = Signal(QRegularExpressionValidator.State)

    def validate(self, input_, pos):
        state, input_, pos = super().validate(input_, pos)
        self.validationChanged.emit(state)
        return state, input_, pos


class ModelZoo(DefaultTab):
    def __init__(self, root, parent, h1_description):
        super().__init__(root, parent, h1_description)
        self._val_pattern = QRegularExpression(r"(\d{3,5},\s*)+\d{3,5}")
        self._set_page()
        self.root.engine_change.connect(self._on_engine_change)
        self.root.engine_change.connect(self._update_available_models)
        self._update_pose_models(self.model_combo.currentText())
        self._update_detectors(self.model_combo.currentText())
        self._destfolder = None
        self.worker = None
        self.thread = None

    @property
    def files(self):
        return self.video_selection_widget.files

    def _set_page(self):
        # Create Run button first so it exists for any method that references it
        self.run_button = QtWidgets.QPushButton("Run")
        self.run_button.setStyleSheet(
            """
            QPushButton {
                background-color: #4CAF50;
                color: white;
                font-weight: bold;
            }
            QPushButton:disabled {
                background-color: #9E9E9E;
                color: white;
                font-weight: bold;
            }
        """
        )
        self.run_button.setFixedWidth(120)
        self.run_button.clicked.connect(self.run_video_inference_superanimal)
        button_layout = QtWidgets.QHBoxLayout()
        button_layout.addStretch()
        button_layout.addWidget(self.run_button)
        button_layout.addStretch()

        self.main_layout.addWidget(_create_label_widget("Video Selection", "font:bold"))
        self.video_selection_widget = VideoSelectionWidget(
            self.root, self, hide_videotype=True
        )
        self.main_layout.addWidget(self.video_selection_widget)

        self._build_common_attributes()
        self._build_tf_attributes()
        self._build_torch_attributes()

        self.home_button = QtWidgets.QPushButton("Return to Welcome page")
        self.home_button.clicked.connect(self.root._generate_welcome_page)
        self.main_layout.addWidget(self.home_button, alignment=Qt.AlignLeft)
        self.help_button = QtWidgets.QPushButton("Help")
        self.help_button.clicked.connect(self.show_help_dialog)
        self.main_layout.addWidget(self.help_button, alignment=Qt.AlignLeft)

        self.go_to_button = QtWidgets.QPushButton("Read Documentation")
        # go to url https://deeplabcut.github.io/DeepLabCut/docs/ModelZoo.html#about-the-superanimal-models when button is clicked
        self.go_to_button.clicked.connect(
            lambda: webbrowser.open(
                "https://deeplabcut.github.io/DeepLabCut/docs/ModelZoo.html#about-the-superanimal-models"
            )
        )
        self.main_layout.addWidget(self.go_to_button, alignment=Qt.AlignLeft)

        # Add the Run button layout
        self.main_layout.addLayout(button_layout)

        self._on_engine_change(self.root.engine)

    def _add_supermodel_section(self, layout: QtWidgets.QGridLayout) -> None:
        # --- Supermodel selection ---
        section_title = QtWidgets.QLabel("Supermodel settings")
        section_title.setStyleSheet("font-weight: bold; font-size: 16px;")
        model_combo_text = QtWidgets.QLabel("Supermodel")
        model_combo_text.setMinimumWidth(150)
        self.model_combo = QtWidgets.QComboBox()
        self.model_combo.setMinimumWidth(250)
        layout.addWidget(section_title, 0, 0, 1, 6)
        layout.addWidget(model_combo_text, 1, 0)
        layout.addWidget(self.model_combo, 1, 1)

    def _add_pose_model_settings_row(self, layout: QtWidgets.QGridLayout):
        # --- Pose Model Type and Pose Confidence Threshold on the same line (now row 2) ---
        pose_model_row = QtWidgets.QHBoxLayout()
        pose_model_label = QtWidgets.QLabel("Pose Model Type")
        pose_model_label.setMinimumWidth(150)
        self.net_type_selector = QtWidgets.QComboBox()
        self.net_type_selector.setMinimumWidth(180)
        pose_conf_label = QtWidgets.QLabel("Pose confidence threshold")
        pose_conf_label.setMinimumWidth(170)
        self.pose_threshold_spinbox = QtWidgets.QDoubleSpinBox(
            decimals=2,
            minimum=0.0,
            maximum=1.0,
            singleStep=0.01,
            value=0.4,
            wrapping=True,
        )
        self.pose_threshold_spinbox.setMaximumWidth(100)
        batch_size_combo_label = QtWidgets.QLabel("Pose model batch size")
        self.batch_size_combo = QtWidgets.QComboBox()
        self.batch_size_combo.setMinimumWidth(100)
        self.batch_size_combo.addItems([str(2**i) for i in range(6)])
        self.batch_size_combo.setCurrentIndex(0)
        pose_model_row.addWidget(pose_model_label)
        pose_model_row.addWidget(self.net_type_selector)
        pose_model_row.addSpacing(20)
        pose_model_row.addWidget(pose_conf_label)
        pose_model_row.addWidget(self.pose_threshold_spinbox)
        pose_model_row.addSpacing(20)
        pose_model_row.addWidget(batch_size_combo_label)
        pose_model_row.addWidget(self.batch_size_combo)
        pose_model_row.addStretch()
        layout.addLayout(pose_model_row, 2, 0, 1, 6)

    def _add_detector_settings_row(self, layout: QtWidgets.QGridLayout):
        # --- Detector Type and Detector Confidence Threshold on the same line (now row 3) ---
        detector_label = QtWidgets.QLabel("Detector Type")
        detector_label.setMinimumWidth(150)
        self.detector_type_selector = QtWidgets.QComboBox()
        self.detector_type_selector.setMinimumWidth(180)
        detector_conf_label = QtWidgets.QLabel("Detector confidence threshold")
        detector_conf_label.setMinimumWidth(170)
        self.detector_threshold_spinbox = QtWidgets.QDoubleSpinBox(
            decimals=2,
            minimum=0.0,
            maximum=1.0,
            singleStep=0.01,
            value=0.1,
            wrapping=True,
        )
        self.detector_threshold_spinbox.setMaximumWidth(100)
        max_individuals_label = QtWidgets.QLabel("Maximum number of individuals")
        max_individuals_label.setMinimumWidth(180)
        self.max_individuals_spinbox = QtWidgets.QSpinBox()
        self.max_individuals_spinbox.setRange(1, 100)
        self.max_individuals_spinbox.setValue(1)
        self.max_individuals_spinbox.setMaximumWidth(100)
        detector_batch_size_combo_label = QtWidgets.QLabel("Detector batch size")
        self.detector_batch_size_combo = QtWidgets.QComboBox()
        self.detector_batch_size_combo.setMinimumWidth(100)
        self.detector_batch_size_combo.addItems([str(2**i) for i in range(6)])
        self.detector_batch_size_combo.setCurrentIndex(0)
        self.detector_row = QtWidgets.QHBoxLayout()
        self.detector_row.addWidget(detector_label)
        self.detector_row.addWidget(self.detector_type_selector)
        self.detector_row.addSpacing(20)
        self.detector_row.addWidget(detector_conf_label)
        self.detector_row.addWidget(self.detector_threshold_spinbox)
        self.detector_row.addSpacing(20)
        self.detector_row.addWidget(max_individuals_label)
        self.detector_row.addWidget(self.max_individuals_spinbox)
        self.detector_row.addSpacing(20)
        self.detector_row.addWidget(detector_batch_size_combo_label)
        self.detector_row.addWidget(self.detector_batch_size_combo)
        self.detector_row.addStretch()
        layout.addLayout(self.detector_row, 3, 0, 1, 6)

    def _add_output_settings_section(self, layout: QtWidgets.QGridLayout):
        loc_label = ClickableLabel("Folder to store results:", parent=self)
        loc_label.signal.connect(self.select_folder)
        self.loc_line = QtWidgets.QLineEdit(
            "<Select a folder - Default: store in same folder as video>",
            self,
        )
        self.loc_line.setReadOnly(True)
        action = self.loc_line.addAction(
            QIcon(os.path.join(BASE_DIR, "assets", "icons", "open2.png")),
            QtWidgets.QLineEdit.TrailingPosition,
        )
        action.triggered.connect(self.select_folder)

        self.create_labeled_video_checkbox = QtWidgets.QCheckBox("Create labeled video")
        self.create_labeled_video_checkbox.setChecked(True)

        layout.addWidget(loc_label, 4, 0)
        layout.addWidget(self.loc_line, 4, 1)
        layout.addWidget(self.create_labeled_video_checkbox, 5, 0)

    def _build_common_attributes(self) -> None:
        settings_layout = _create_grid_layout(margins=(20, 0, 0, 0))

        self._add_supermodel_section(settings_layout)
        self._add_pose_model_settings_row(settings_layout)
        self._add_detector_settings_row(settings_layout)
        self._add_output_settings_section(settings_layout)

        self.settings_widget = QtWidgets.QWidget()
        self.settings_widget.setLayout(settings_layout)
        self.main_layout.addWidget(self.settings_widget)

        self.model_combo.currentTextChanged.connect(self._update_pose_models)
        self.model_combo.currentTextChanged.connect(self._update_detectors)
        self.model_combo.currentTextChanged.connect(
            self._update_adaptation_detector_visibility
        )

    def _add_tf_scales_row(self, layout: QtWidgets.QGridLayout):
        scales_label = QtWidgets.QLabel("Scale list")
        scales_label.setMinimumWidth(300)
        self.scales_line = QtWidgets.QLineEdit("", parent=self)
        self.scales_line.setMinimumWidth(500)
        self.scales_line.setPlaceholderText(
            "Optionally input a list of integer sizes separated by commas..."
        )
        validator = RegExpValidator(self._val_pattern, self)
        validator.validationChanged.connect(self._handle_validation_change)
        self.scales_line.setValidator(validator)
        tooltip_label = QtWidgets.QLabel()
        tooltip_label.setPixmap(
            QPixmap(
                os.path.join(BASE_DIR, "assets", "icons", "help2.png")
            ).scaledToWidth(30)
        )
        tooltip_label.setToolTip(
            "Approximate animal sizes in pixels, for spatial pyramid search. If left "
            "blank, defaults to video height +/- 50 pixels"
        )
        scales_row = QtWidgets.QHBoxLayout()
        scales_row.addWidget(scales_label)
        scales_row.addWidget(self.scales_line)
        scales_row.addWidget(tooltip_label)
        layout.addLayout(scales_row, 1, 0, 1, 2)

    def _add_use_adaptation_row(self, layout: QtWidgets.QGridLayout, layout_row: int):
        # --- Adaptation Checkbox with Help Button (TF section) ---
        self.adapt_checkbox = QtWidgets.QCheckBox("Use video adaptation")
        self.adapt_checkbox.setChecked(True)
        self.adapt_checkbox.setStyleSheet(
            "font-weight: bold; font-size: 16px; padding: 6px 12px;"
        )
        # Add help button
        adapt_help_btn = QtWidgets.QToolButton()
        adapt_help_btn.setIcon(
            QIcon(os.path.join(BASE_DIR, "assets", "icons", "help2.png"))
        )
        adapt_help_btn.setIconSize(QSize(24, 24))
        adapt_help_btn.setToolTip("What is video adaptation?")

        def show_adapt_help():
            QtWidgets.QMessageBox.information(
                self,
                "Video Adaptation",
                "This will adapt the model on the fly to your video data in a self-supervised way.",
            )

        adapt_help_btn.clicked.connect(show_adapt_help)
        use_adaptation_row = QtWidgets.QHBoxLayout()
        use_adaptation_row.addWidget(self.adapt_checkbox)
        use_adaptation_row.addWidget(adapt_help_btn)
        use_adaptation_row.addStretch()
        layout.addLayout(use_adaptation_row, layout_row, 0, 1, 2)

    def _add_tf_adaptation_settings_row(self, layout: QtWidgets.QGridLayout):
        pseudo_threshold_label = QtWidgets.QLabel("Pseudo-label confidence threshold")
        self.pseudo_threshold_spinbox = QtWidgets.QDoubleSpinBox(
            decimals=2,
            minimum=0.01,
            maximum=1.0,
            singleStep=0.05,
            value=0.1,
            wrapping=True,
        )
        self.pseudo_threshold_spinbox.setMaximumWidth(100)
        adapt_iter_label = QtWidgets.QLabel("Number of adaptation iterations")
        adapt_iter_label.setMinimumWidth(300)
        self.adapt_iter_spinbox = QtWidgets.QSpinBox()
        self.adapt_iter_spinbox.setRange(100, 10000)
        self.adapt_iter_spinbox.setValue(1000)
        self.adapt_iter_spinbox.setSingleStep(100)
        self.adapt_iter_spinbox.setGroupSeparatorShown(True)
        self.adapt_iter_spinbox.setMaximumWidth(300)
        self.tf_adaptation_settings_row = QtWidgets.QHBoxLayout()
        self.tf_adaptation_settings_row.addWidget(pseudo_threshold_label)
        self.tf_adaptation_settings_row.addWidget(self.pseudo_threshold_spinbox)
        self.tf_adaptation_settings_row.addSpacing(20)
        self.tf_adaptation_settings_row.addWidget(adapt_iter_label)
        self.tf_adaptation_settings_row.addWidget(self.adapt_iter_spinbox)
        layout.addLayout(self.tf_adaptation_settings_row, 3, 0, 1, 6)

    def _build_tf_attributes(self) -> None:
        tf_settings_layout = _create_grid_layout(margins=(20, 0, 0, 0))

        self._add_tf_scales_row(tf_settings_layout)
        self._add_use_adaptation_row(tf_settings_layout, 2)
        self._add_tf_adaptation_settings_row(tf_settings_layout)

        self.adapt_checkbox.stateChanged.connect(self._adapt_checkbox_status_changed)

        self.tf_widget = QtWidgets.QWidget()
        self.tf_widget.setLayout(tf_settings_layout)
        self.tf_widget.hide()
        self.main_layout.addWidget(self.tf_widget)

    def _add_torch_adaptation_settings_row(self, layout: QtWidgets.QGridLayout):
        # Compact adaptation settings row
        pseudo_threshold_label = QtWidgets.QLabel("Pseudo-label confidence threshold")
        pseudo_threshold_label.setMinimumWidth(200)
        self.torch_pseudo_threshold_spinbox = QtWidgets.QDoubleSpinBox(
            decimals=2,
            minimum=0.01,
            maximum=1.0,
            singleStep=0.05,
            value=0.1,
            wrapping=True,
        )
        self.torch_pseudo_threshold_spinbox.setMaximumWidth(100)
        adapt_epoch_label = QtWidgets.QLabel("Number of adaptation epochs")
        adapt_epoch_label.setMinimumWidth(180)
        self.torch_adapt_epoch_spinbox = QtWidgets.QSpinBox()
        self.torch_adapt_epoch_spinbox.setRange(1, 50)
        self.torch_adapt_epoch_spinbox.setValue(4)
        self.torch_adapt_epoch_spinbox.setMaximumWidth(100)
        self.adapt_det_epoch_label = QtWidgets.QLabel(
            "Number of detector adaptation epochs"
        )
        self.adapt_det_epoch_label.setMinimumWidth(200)
        self.torch_adapt_det_epoch_spinbox = QtWidgets.QSpinBox()
        self.torch_adapt_det_epoch_spinbox.setRange(1, 50)
        self.torch_adapt_det_epoch_spinbox.setValue(4)
        self.torch_adapt_det_epoch_spinbox.setMaximumWidth(100)
        self.torch_adaptation_settings_row = QtWidgets.QHBoxLayout()
        self.torch_adaptation_settings_row.addWidget(pseudo_threshold_label)
        self.torch_adaptation_settings_row.addWidget(
            self.torch_pseudo_threshold_spinbox
        )
        self.torch_adaptation_settings_row.addSpacing(20)
        self.torch_adaptation_settings_row.addWidget(adapt_epoch_label)
        self.torch_adaptation_settings_row.addWidget(self.torch_adapt_epoch_spinbox)
        self.torch_adaptation_settings_row.addSpacing(20)
        self.torch_adaptation_settings_row.addWidget(self.adapt_det_epoch_label)
        self.torch_adaptation_settings_row.addWidget(self.torch_adapt_det_epoch_spinbox)
        self.torch_adaptation_settings_row.addStretch()
        layout.addLayout(self.torch_adaptation_settings_row, 2, 0, 1, 6)

    def _build_torch_attributes(self) -> None:
        torch_settings_layout = _create_grid_layout(margins=(20, 0, 0, 0))

        self._add_use_adaptation_row(torch_settings_layout, 1)
        self._add_torch_adaptation_settings_row(torch_settings_layout)

        self.adapt_checkbox.stateChanged.connect(self._adapt_checkbox_status_changed)

        self.torch_widget = QtWidgets.QWidget()
        self.torch_widget.setLayout(torch_settings_layout)
        self.torch_widget.hide()
        self.main_layout.addWidget(self.torch_widget)

    def _adapt_checkbox_status_changed(self, state: int) -> None:
        if self.root.engine == Engine.TF:
            set_layout_contents_visible(
                self.tf_adaptation_settings_row, Qt.CheckState(state) == Qt.Checked
            )
        elif self.root.engine == Engine.PYTORCH:
            set_layout_contents_visible(
                self.torch_adaptation_settings_row, Qt.CheckState(state) == Qt.Checked
            )
            if Qt.CheckState(state) == Qt.Checked:
                self._update_adaptation_detector_visibility(
                    self.model_combo.currentText()
                )

    def select_folder(self):
        dirname = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Please select a folder", self.root.project_folder
        )
        if not dirname:
            return

        self._destfolder = dirname
        self.loc_line.setText(dirname)

    def show_help_dialog(self):
        dialog = QtWidgets.QDialog(self)
        layout = QtWidgets.QVBoxLayout()
        help_text = deeplabcut.video_inference_superanimal.__doc__

        label = QtWidgets.QLabel(help_text, self)
        scroll = QtWidgets.QScrollArea()
        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        scroll.setWidgetResizable(True)
        scroll.setWidget(label)
        layout.addWidget(scroll)
        dialog.setLayout(layout)
        dialog.exec_()

    def _handle_validation_change(self, state):
        if state == RegExpValidator.Invalid:
            color = "red"
        elif state == RegExpValidator.Intermediate:
            color = "gold"
        elif state == RegExpValidator.Acceptable:
            color = "lime"
        self.scales_line.setStyleSheet(f"border: 1px solid {color}")
        QTimer.singleShot(500, lambda: self.scales_line.setStyleSheet(""))

    def run_video_inference_superanimal(self):
        files = list(self.files)
        if not files:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("You must select video files")
            msg.setWindowTitle("Error")
            msg.setMinimumWidth(400)
            msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
            msg.exec_()
            return

        supermodel_name = self.model_combo.currentText()
        create_labeled_video = self.create_labeled_video_checkbox.isChecked()
        batch_size = int(self.batch_size_combo.currentText())
        detector_batch_size = int(self.detector_batch_size_combo.currentText())
        kwargs = self._gather_kwargs()

        can_run_in_background = True
        self.run_button.setEnabled(False)
        self.root._progress_bar.show()
        try:
            # Use standard function for other models
            if can_run_in_background:
                func = partial(
                    deeplabcut.video_inference_superanimal,
                    files,
                    supermodel_name,
                    dest_folder=self._destfolder,
                    create_labeled_video=create_labeled_video,
                    batch_size=batch_size,
                    detector_batch_size=detector_batch_size,
                    **kwargs,
                )
                self.worker, self.thread = move_to_separate_thread(func)
                self.worker.finished.connect(self.signal_analysis_complete)
                self.thread.start()
            else:
                print(f"Calling video_inference_superanimal with kwargs={kwargs}")
                results = deeplabcut.video_inference_superanimal(
                    files,
                    supermodel_name,
                    dest_folder=self._destfolder,
                    create_labeled_video=create_labeled_video,
                    batch_size=batch_size,
                    detector_batch_size=detector_batch_size,
                    **kwargs,
                )
                self.signal_analysis_complete()

        except Exception as e:
            print(f"[Error] {e}")
            self.run_button.setEnabled(True)
            self.root._progress_bar.hide()

    def signal_analysis_complete(self):
        self.run_button.setEnabled(True)
        self.root._progress_bar.hide()

        # Check if labeled videos were actually created
        files = list(self.files)
        videos_created = []

        # Determine the output folder
        output_folder = self._destfolder if self._destfolder else Path(files[0]).parent

        for video_path in files:
            video_name = Path(video_path).stem
            labeled_videos = list(
                Path(output_folder).glob(f"{video_name}_*_labeled*.mp4")
            )
            if labeled_videos:
                videos_created.extend([str(v) for v in labeled_videos])

        # Show appropriate message
        if videos_created:
            msg = QtWidgets.QMessageBox(
                text=f"SuperAnimal video inference complete!\n\nCreated labeled videos:\n"
                + "\n".join(videos_created)
            )
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.exec_()
        else:
            msg = QtWidgets.QMessageBox(
                text=f"SuperAnimal video inference complete, but no labeled videos were created."
            )
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.exec_()

    def stop_processes(self):
        """Stop any running processes"""
        if self.thread and self.thread.isRunning():
            print("Stopping running processes...")
            self.thread.quit()
            self.thread.wait(5000)  # Wait up to 5 seconds
            if self.thread.isRunning():
                self.thread.terminate()
                self.thread.wait(2000)
            self.worker = None
            self.thread = None
            self.run_button.setEnabled(True)
            self.root._progress_bar.hide()

    def closeEvent(self, event):
        """Override closeEvent to stop processes when tab is closed"""
        self.stop_processes()
        super().closeEvent(event)

    def _gather_kwargs(self) -> dict:
        kwargs = dict(model_name=self.net_type_selector.currentText())

        if self.root.engine == Engine.TF:
            scales = []
            scales_ = self.scales_line.text()
            if scales_:
                if (
                    self.scales_line.validator().validate(scales_, 0)[0]
                    == RegExpValidator.Acceptable
                ):
                    scales = list(map(int, scales_.split(",")))
            kwargs["scale_list"] = scales
            kwargs["video_adapt"] = self.adapt_checkbox.isChecked()
            kwargs["pseudo_threshold"] = self.pseudo_threshold_spinbox.value()
            kwargs["adapt_iterations"] = self.adapt_iter_spinbox.value()
        else:
            kwargs["detector_name"] = self.detector_type_selector.currentText()
            kwargs["video_adapt"] = (self.adapt_checkbox.isChecked())
            kwargs["pseudo_threshold"] = self.pose_threshold_spinbox.value()
            kwargs["bbox_threshold"] = self.detector_threshold_spinbox.value()
            kwargs["detector_epochs"] = self.torch_adapt_det_epoch_spinbox.value()
            kwargs["pose_epochs"] = self.torch_adapt_epoch_spinbox.value()
            kwargs["max_individuals"] = self.max_individuals_spinbox.value()

        return kwargs

    def _update_available_models(self, engine: Engine) -> None:
        current_dataset = self.model_combo.currentText()

        if engine == Engine.TF:
            supermodels = ["superanimal_topviewmouse", "superanimal_quadruped"]
        else:
            supermodels = dlclibrary.get_available_datasets()

        set_combo_items(
            combo_box=self.model_combo,
            items=supermodels,
            index=(
                supermodels.index(current_dataset)
                if current_dataset in supermodels
                else 0
            ),
        )

    def _update_pose_models(self, super_animal: str) -> None:
        if len(super_animal) == 0:
            set_combo_items(combo_box=self.net_type_selector, items=[])
            return

        set_combo_items(
            combo_box=self.net_type_selector,
            items=(
                ["dlcrnet"]
                if self.root.engine == Engine.TF
                else dlclibrary.get_available_models(super_animal)
            ),
        )

    def _update_detectors(self, super_animal: str) -> None:
        if len(super_animal) == 0:
            set_combo_items(combo_box=self.detector_type_selector, items=[])
            return

        items = []
        if self.root.engine == Engine.PYTORCH:
            if super_animal == "superanimal_humanbody":
                items = list(TORCHVISION_DETECTORS.keys())
            else:
                items = dlclibrary.get_available_detectors(super_animal)
        set_combo_items(combo_box=self.detector_type_selector, items=items)
        set_layout_contents_visible(
            self.detector_row, self.root.engine == Engine.PYTORCH
        )

    def _update_adaptation_detector_visibility(self, superanimal: str):
        self.adapt_det_epoch_label.setVisible((superanimal != "superanimal_humanbody"))
        self.torch_adapt_det_epoch_spinbox.setVisible(
            (superanimal != "superanimal_humanbody")
        )

    @Slot(Engine)
    def _on_engine_change(self, engine: Engine) -> None:
        self._update_available_models(engine)
        if engine == Engine.PYTORCH:
            self.tf_widget.hide()
            self.torch_widget.show()
        else:
            self.torch_widget.hide()
            self.tf_widget.show()

        # Hide widgets in detector row
        set_layout_contents_visible(self.detector_row, engine == Engine.PYTORCH)



#
# DeepLabCut Toolbox (deeplabcut.org)
# © A. & M.W. Mathis Labs
# https://github.com/DeepLabCut/DeepLabCut
#
# Please see AUTHORS for contributors.
# https://github.com/DeepLabCut/DeepLabCut/blob/master/AUTHORS
#
# Licensed under GNU Lesser General Public License v3.0
#

from deeplabcut.benchmark.cli import main

if __name__ == "__main__":
    main()
